/* codeslots.c
 * Code slots.
 *
 * Copyright (c) 1996 T. J. Wilkinson & Associates, London, UK.
 *
 * See the file "license.terms" for information on usage and
 * redistribution of this file.
 *
 * Written by Tim Wilkinson <tim@tjwassoc.demon.co.uk>
 */

#define	FDBG(s)

#include "config.h"
#include "config-std.h"
#include "gtypes.h"
#include "slots.h"
#include "icode.h"
#include "icode-common.h"
#include "flows.h"
#include "code-analyse.h"
#include "md.h"

static void _free_reg(slotreg*);
static void _attach_reg_to_slot(SlotInfo*, slotreg*);
static void _detach_reg_from_slot(SlotInfo*, slotreg*);
static slotreg* _alloc_reg(int);

extern void spill(SlotInfo*);
extern void reload(SlotInfo*);

extern int maxArgs;
extern int maxLocal;
extern int maxStack;
extern int maxTemp;
extern int isStatic;

#define	ADD_REGISTER(_type, _regno, _flags)		\
	(ptr)->regno = (_regno);			\
	(ptr)->myslot = 0;				\
	(ptr)->modified = 0;				\
	(ptr)->type = (_type);				\
	(ptr)->next = registerTypes[(_type)].free;	\
	registerTypes[(_type)].free = (ptr);		\
	(ptr)++

static int translateType[MAX_TRANSLATION_TYPES] = SLOTTYPES;
static typedata registerTypes[MAX_REGISTER_TYPES];

static int number_of_SlotInfo = -1;

static SlotInfo* basicSlotInfo;
SlotInfo* cslotinfo;
SlotInfo* localinfo;
SlotInfo* stackinfo;
SlotInfo* tempinfo;
SlotInfo* returninfo;
activeState activeInner;

slotreg registerData[MAX_REGISTERS];

/*
 * Initialise the slot+register allocation scheme.
 */
void
initCodeSlots(int nrSlotInfo)
{
	int i;
	slotreg* ptr;

	for (i = 0; i < MAX_REGISTER_TYPES; i++) {
		registerTypes[i].free = 0;
		registerTypes[i].inuse = 0;
	}

	ptr = &registerData[0];

	/* Initiate registers */
	INIT_REGISTERS();

	/* Check we've used every register up */
	assert(ptr == &registerData[MAX_REGISTERS]);

	/* Allocate the SlotInfo if necessary */
	if (nrSlotInfo > number_of_SlotInfo) {
		if (basicSlotInfo != 0) {
			free(basicSlotInfo);
		}
		basicSlotInfo = calloc((nrSlotInfo + MAXRETURNS), sizeof(SlotInfo));
		assert(basicSlotInfo != 0);

		activeInner.flow = 0;

		number_of_SlotInfo = nrSlotInfo;
	}

	memset(basicSlotInfo, 0, (number_of_SlotInfo + MAXRETURNS) * sizeof(SlotInfo));
	cslotinfo = &basicSlotInfo[MAXRETURNS];
	returninfo = &basicSlotInfo[0];

	/* Setup return SlotInfo */
	INIT_RETURNSLOTS();
}

/*
 * Free a slot up.
 */
void
freeSlot(SlotInfo* myslot)
{
	_free_reg(myslot->reg);
	_detach_reg_from_slot(myslot, myslot->reg);
}

/*
 * Allocate a register to a slot and reload its value.  The issuenr is
 * used to prevent register reuse in the same instruction.
 */
int
demandRegister(SlotInfo* myslot, int op, int type, uint32 issuenr)
{
	slotreg* myreg;
	int rtype;

	/* If I already have a register and it's the correct type, reuse it.
	 * If not, spill it (if we won't overwrite the space) and free the
	 * register up.
	 */
	rtype = translateType[type];
	if (myslot->reg != 0) {
		if (myslot->reg->type == rtype) {
			myslot->reg->issuenr = issuenr;
			myslot->reg->modified |= op;
			myslot->reg->ctype = type;
			return (myslot->reg->regno);
		}

		/* Return SlotInfo should not get here */
		assert(slot_return(myslot) == 0);

		/* We cannot assert this earlier because 'myslot' may be
		 * a return slot register and this will fail the assert.
		 */
		assert(myslot->reg->myslot == myslot);
		if (op & rread) {
			spill(myslot);
			if (myslot->fixedreg != 0) {
				activeInner.flow->cancel = true;
			}
		}
		_free_reg(myslot->reg);
		_detach_reg_from_slot(myslot, myslot->reg);
	}

	/* Return SlotInfo should not get here */
	assert(slot_return(myslot) == 0);

	/* If type is Rlong or Rdouble then we must clobber the following
	 * slot in case it's in use.
	 */
	if ((type == Rlong || type == Rdouble) && myslot[1].reg != 0) {
		if (op & rread) {
			spill(&myslot[1]);
			if (myslot[1].fixedreg != 0) {
				activeInner.flow->cancel = true;
			}
		}
		_free_reg(myslot[1].reg);
		_detach_reg_from_slot(&myslot[1], myslot[1].reg);
	}

	/* Select a register.  If we attempt to reload a fixed register,
	 * cancel the flow and let the engine restart code generation.
	 */
	myreg = _alloc_reg(type);
	assert(myreg != 0);

	myreg->ctype = type;
	assert(myreg->issuenr <= issuenr);
	myreg->issuenr = issuenr;

	_attach_reg_to_slot(myslot, myreg);
	if (op & rread) {
		reload(myslot);
		if (myslot->fixedreg != 0) {
			activeInner.flow->cancel = true;
		}
	}
	myreg->modified = op;

	assert(myslot->reg->myslot == myslot);

	return (myreg->regno);
}

/*
 * Force a particular register to a slot.
 */
void
forceRegister(SlotInfo* myslot, slotreg* rptr, int type, uint32 issuenr)
{
	assert(type >= 0 && type < MAX_TRANSLATION_TYPES);
	assert(translateType[type] != -1);
	assert(translateType[type] == rptr->type);

	if (rptr != myslot->reg) {
		slotreg* r = myslot->reg;
		_free_reg(myslot->reg);
		_detach_reg_from_slot(myslot, myslot->reg);
		_detach_reg_from_slot(rptr->myslot, rptr);
		_attach_reg_to_slot(myslot, rptr);
		rptr->modified = rwrite;
		assert(r == 0 || r->myslot == 0);
	}
	rptr->ctype = type;
	assert(rptr->issuenr <= issuenr);
	rptr->issuenr = issuenr;

	assert(myslot->reg->myslot == myslot);
}

/*
 * Change the register associated with a slot, loading, saving and
 * moving the contents as necessary.
 */
void
changeRegister(SlotInfo* myslot, slotreg* rptr, int type, uint32 issuenr)
{
	/* If the desired register is in use, spill it. Then detach
	 * it from slot or freelist.
	 */
	if (rptr->myslot != 0) {
		spill(rptr->myslot);
		if (rptr->myslot->fixedreg != 0) {
			activeInner.flow->cancel = true;
		}
	}
	_detach_reg_from_slot(rptr->myslot, rptr);

	/* Change register type to what we want. */
	assert(rptr->type == translateType[type]);
	rptr->ctype = type;

	/* If we don't have a register associated with my slot, rebind it */
	if (myslot->reg == 0) {
		_attach_reg_to_slot(myslot, rptr);
		reload(myslot);
		if (myslot->fixedreg != 0) {
			activeInner.flow->cancel = true;
		}
	}
	/* If I do (and it's not my current one), transfer the contents
	 * to the new register and free the old one.
	 */
	else if (myslot->reg != rptr) {
		slotreg* r = myslot->reg;

		assert(rptr->issuenr < issuenr);

		spill(myslot);
		if (myslot->fixedreg != 0) {
			activeInner.flow->cancel = true;
		}
		_free_reg(myslot->reg);
		_detach_reg_from_slot(myslot, myslot->reg);

		_attach_reg_to_slot(myslot, rptr);
		reload(myslot);
		if (myslot->fixedreg != 0) {
			activeInner.flow->cancel = true;
		}

		assert(r->myslot == 0);
	}
	rptr->issuenr = issuenr;

	assert(myslot->reg->myslot == myslot);
}

/*
 * Clobber a particular register.
 */
void
clobberRegister(slotreg* rptr)
{
	/* If the desired register is in use, spill it, detach
	 * it from slot and free it.
	 */
	if (rptr->myslot != 0) {
		spill(rptr->myslot);
		if (rptr->myslot->fixedreg != 0) {
			activeInner.flow->cancel = true;
		}
		_detach_reg_from_slot(rptr->myslot, rptr);
		_free_reg(rptr);
	}
}

/*
 * Force a register back into its memory slot.  If the register is live
 * and modified, make sure we write it back first.
 */
int
demandOffset(SlotInfo* myslot, int op)
{
	if (myslot->reg != 0) {
		if (op & rread) {
			spill(myslot);
			if (myslot->fixedreg != 0) {
				activeInner.flow->cancel = true;
			}
		}
		_free_reg(myslot->reg);
		_detach_reg_from_slot(myslot, myslot->reg);
	}
	myslot->offset = SLOT2FRAMEOFFSET(myslot);

	assert(myslot->reg == 0);

	return (myslot->offset);
}

/* ------------------------------------------------------------------------- */

void
activateInner(innerLoop* flow, int issuenr)
{
	slotreg* rptr;
	int i;

	/* Make sure we can activate this flow */
	assert(activeInner.flow == 0);

	activeInner.flow = flow;

	/* Setup new masks */
	for (i = 0; i < codeInfo.slotlen; i++) {
		if (flow->slot[i] != 0) {
			rptr = cslotinfo[i].reg;
			if (rptr == 0) {
				rptr = _alloc_reg(Rint);
				rptr->ctype = Rint;
				assert(rptr->issuenr <= issuenr);
				_attach_reg_to_slot(&cslotinfo[i], rptr);
				reload(&cslotinfo[i]);
				rptr->modified = rread|rwrite;
			}
			cslotinfo[i].fixedreg = rptr;
			rptr->issuenr = issuenr;
		}
	}
}

void
deactivateInner(void)
{
	int i;

	for (i = 0; i < codeInfo.slotlen; i++) {
		cslotinfo[i].fixedreg = 0;
	}
	activeInner.flow = 0;
}

/* ------------------------------------------------------------------------- */

/*
 * Allocate a register of the specified type from the freelist if possible,
 * otherwise pick the least recently used one.
 */
static
slotreg*
_alloc_reg(int ctype)
{
	slotreg* rptr;
	slotreg* srptr;
	int type;

	/* Translate type to internal type */
	assert(ctype >= 0 && ctype < MAX_TRANSLATION_TYPES);
	assert(translateType[ctype] != -1);
	type = translateType[ctype];

	/* Allocate from the freelist if possible */
	rptr = registerTypes[type].free;
	if (rptr != 0) {
		registerTypes[type].free = rptr->next;
		rptr->next = registerTypes[type].inuse;
		registerTypes[type].inuse = rptr;
		assert(rptr->myslot == 0);
		return (rptr);
	}

	/* Otherwise, select the least recently used register */
	srptr = registerTypes[type].inuse;
	assert(srptr != 0);
	for (rptr = srptr->next; rptr != 0; rptr = rptr->next) {
		if (rptr->issuenr < srptr->issuenr) {
			srptr = rptr;
		}
	}
	spill(srptr->myslot);
	if (srptr->myslot->fixedreg != 0) {
		activeInner.flow->cancel = true;
	}
	srptr->myslot->reg = 0;
	srptr->myslot = 0;
	return (srptr);
}

/*
 * Freeup a register and return it to the freelist.
 */
static
void
_free_reg(slotreg* myreg)
{
	slotreg* rptr;

	if (myreg == 0) {
		return;
	}

	/* Remove register from the inuse list */
	rptr = registerTypes[myreg->type].inuse;
	if (rptr == myreg) {
		registerTypes[myreg->type].inuse = myreg->next;
	}
	else for (; rptr->next != 0; rptr = rptr->next) {
		if (rptr->next == myreg) {
			rptr->next = myreg->next;
			break;
		}
	}

	/* Attach to free list */
	myreg->next = registerTypes[myreg->type].free;
	registerTypes[myreg->type].free = myreg;
}

/*
 * Attach a register to a slot.
 */
static
void
_attach_reg_to_slot(SlotInfo* myslot, slotreg* myreg)
{
	assert(myslot->reg == 0);
	assert(myreg->myslot == 0);

	myslot->reg = myreg;
	myreg->myslot = myslot;
}

/*
 * Detach a register from a slot.
 */
static
void
_detach_reg_from_slot(SlotInfo* myslot, slotreg* myreg)
{
	slotreg* rptr;

	if (myslot == 0) {
		assert(myreg != 0);
		assert(myreg->myslot == 0);

		/* Detach register from free list */
		rptr = registerTypes[myreg->type].free;
		if (rptr == myreg) {
			registerTypes[myreg->type].free = rptr->next;
		}
		else for (; rptr->next != 0; rptr = rptr->next) {
			if (rptr->next == myreg) {
				rptr->next = myreg->next;
				break;
			}
		}

		/* Attach to inuse list */
		myreg->next = registerTypes[myreg->type].inuse;
		registerTypes[myreg->type].inuse = myreg;
	}
	else {
		assert(myslot->reg == myreg);
		if (myreg != 0) {
			assert(myreg->myslot == myslot);
			myslot->reg = 0;
			myreg->modified = 0;
			myreg->myslot = 0;
		}
	}
}
