/* machine.c
 * Translate the Kaffe instruction set to the native one.
 *
 * Copyright (c) 1996 T. J. Wilkinson & Associates, London, UK.
 *
 * See the file "license.terms" for information on usage and
 * redistribution of this file.
 *
 * Written by Tim Wilkinson <tim@tjwassoc.demon.co.uk>
 */

#define	DBG(s)
#define	MDBG(s)
#define	FDBG(s)

#include "config.h"
#include "config-std.h"
#include "gtypes.h"
#include "bytecode.h"
#include "slots.h"
#include "machine.h"
#include "icode.h"
#include "icode-common.h"
#include "labels.h"
#include "codeproto.h"
#include "access.h"
#include "object.h"
#include "constants.h"
#include "baseClasses.h"
#include "classMethod.h"
#include "code.h"
#include "access.h"
#include "lookup.h"
#include "exception.h"
#include "gc.h"
#include "flags.h"
#include "external.h"
#include "support.h"
#include "flows.h"
#include "code-analyse.h"
#include "errors.h"
#include "md.h"

/*
 * Define information about this engine.
 */
char* engine_name = "Fast just-in-time";
char* engine_version = KVER;

int stackno;
int maxStack;
int maxLocal;
int maxTemp;
int maxArgs;
int isStatic;

/* Fix to so we can reuse libraries without change */
SlotInfo retarg;

int tmpslot = 0;
int argcount = 0;		/* Function call argument count */
int32 pc;
int32 npc;
int32 ppc;

/* Unit in which code block is allocated */
#define	ALLOCCODEBLOCKSZ	(1024 * 1024)
/* Codeblock redzone - allows for safe overrun when generating instructions */
#define	CODEBLOCKREDZONE	1024

static nativecode* codeblock;
static int codeblock_size;
static nativecode* codeblock_base;
static nativecode* codeblock_end;

/* Mutex to protect the translator from multiple entry */
static iMux translatorMux;

nativecode* CODEPC;

struct {
	int	time;
	int	bytecodes;
} jitStats;

extern activeState activeInner;
extern int issuenr;
extern uint8 insnLen[];

extern void spill(SlotInfo*);
extern void reload(SlotInfo*);

void endBlock(void);
void startBlock(void);
void startInsn(uintp);
void prepareFunctionCall(void);
void fixupFunctionCall(void);

/*
 * Translate a method into native code.
 */
void
translate(methods* meth)
{
	jint low;
	jint high;
	int idx;
	SlotInfo* cnst;
	SlotInfo* tmp1;
	SlotInfo* tmp2;
	SlotInfo* tmp3;
	SlotInfo* tmp4;
	SlotInfo* mtable;

	int do_arrayboundcheck;

	bytecode* base;
	int len;
	int nx;
	int sx;
	int oldstackno;

	uint32 clen;
	int i;
	jexception* e;

	SlotInfo* s0;
	SlotInfo* s1;
	SlotInfo* s2;
	SlotInfo* s3;
	jint v0;

	innerLoop* flow;
	innerLoop* nflow;

	callInfo cinfo;
	Field *field;
	Class* crinfo;
	bool wide;

	int64 tms;
	int64 tme;

	/* Only one in the translator at once. Must check the translation
	 * hasn't been done by someone else once we get it.
	 */ 
	_lockMutex(&translatorMux);
	if (meth->ncode != 0) {
		_unlockMutex(&translatorMux);
		return;
	}

	if (flag_jit) {
		tms = currentTime();
	}

DBG(	printf("callinfo = 0x%x\n", &cinfo);	)

	/* If this code block is native, then just set it up and return */
	if ((meth->accflags & ACC_NATIVE) != 0) {
		native(meth);
		_unlockMutex(&translatorMux);
		return;
	}

	maxLocal = meth->localsz;
	maxStack = meth->stacksz;
	maxArgs = meth->ins;
	if (meth->accflags & ACC_STATIC) {
		isStatic = 1;
	}
	else {
		isStatic = 0;
		maxArgs += 1;
	}

	base = (bytecode*)meth->code;
	len = meth->codelen;

	/* Set locals, stack and temps */
	assert(maxLocal + maxStack < MAXSLOT);
	initCodeSlots(MAXSLOT);
	localinfo = &cslotinfo[0];
	stackinfo = &localinfo[maxLocal];
	tempinfo = &stackinfo[maxStack];

DBG(	printf("localinfo 0x%x stackinfo 0x%x\n", localinfo, stackinfo); )

	/* Clear temporary counter */
	maxTemp = 8;	/* HACK HACK HACK */

	/* Now analyse the bytecode */
	verifyMethod(meth);

	/* Next reduce bytecode to native code */

	if (codeblock_base == 0) {
		restart:
		codeblock_size = ALLOCCODEBLOCKSZ;
		codeblock_base = calloc(sizeof(nativecode), codeblock_size + CODEBLOCKREDZONE);
		assert(codeblock_base != 0);
		codeblock_end = codeblock_base + codeblock_size;
		codeblock = codeblock_base;
	}

	restart_flow:

	initLabels();

	CODEPC = codeblock;
	flow = 0;

	start_function();
	monitor_enter();

	wide = false;
	nx = 0;
	sx = 0;
	ppc = 0;
	start_basic_block();
	stackno = meth->localsz + meth->stacksz;
	for (pc = 0; pc < len; pc = npc) {

		/* If we overrun the codeblock, allocate a new one and
		 * restart.
		 */
		if (CODEPC >= codeblock_end) {
			goto restart;
		}

		if (stackno < meth->localsz || stackno > meth->localsz + meth->stacksz) {
			printf("%d: Stack out out bounds: %d (0-%d)\n", pc, stackno, meth->stacksz);
			fflush(stdout);
			abort();
		}

		/* Reset various optimisation flags */
		do_arrayboundcheck = 1;

		/* Check for any flows to activate */
		nflow = FLOW(pc);
		if (nflow != 0 && flow == 0 && nflow->cancel == false) {
			issuenr++;
			activateInner(nflow, issuenr);
FDBG(				printf("Start flow @ %d:%d (0x%x)\n",
				pc, nflow->in, CODEPC);		)
			flow = nflow;
		}

		npc = pc + insnLen[base[pc]];

		/* Note start of exception handling blocks */
		if (IS_STARTOFEXCEPTION(pc)) {
			stackno = meth->localsz + meth->stacksz - 1;
			start_exception_block();
		}

		oldstackno = stackno;

		/* If this instruction is widened, adjust the next pc
		 * as appropriate.
		 */
		if (wide) {
			switch(base[pc]) {
			case ILOAD:
			case FLOAD:
			case ALOAD:
			case LLOAD:
			case DLOAD:
			case ISTORE:
			case FSTORE:
			case ASTORE:
			case LSTORE:
			case DSTORE:
				npc += 1;
				break;
			case IINC:
				npc += 2;
				break;
			default: 
				fprintf(stderr, "Badly widened instruction %d\n" , base[pc]);            
				throwException(VerifyError);
				break;
			}
		}

		start_instruction();
		switch (base[pc]) {
		default:
			fprintf(stderr, "Unknown bytecode %d\n", base[pc]);
			abort();
#include "kaffe.def"
		}

		/* Note maximum number of temp slots used and reset it */
		if (tmpslot > 0) {
			assert(tmpslot < maxTemp);
			tmpslot = 0;
		}

		skip:

		/* Check to see if this flow has generated any overflow
		 * errors we can't handle.  If it does we deactivate this
		 * flow, invalidate it, and re-run code generation.
		 */
		if (flow != 0 && flow->cancel == true) {
			deactivateInner();
FDBG(			printf("Cancel flow @ %d:%d (0x%x)\n", npc,
				flow->out, CODEPC); fflush(stdout);	)
			goto restart_flow;
		}

		/* Kill any stack slots which are no longer valid */
		for (idx = oldstackno; idx < stackno; idx++) {
			if (localinfo[idx].reg != 0) {
				localinfo[idx].reg->modified = 0;
			}
		}

		/* Check for any flows to deactivate */
		if (flow != 0 && npc >= flow->out) {
			deactivateInner();
FDBG(			printf("Flow @ %d--%d (0x%x)\n", flow->in,
				flow->out, CODEPC);			)
			flow = 0;
		}

		/* If about to start new basic block, end the old one,
		 * start the new one and get the new stack pointer.
		 */
		if (IS_STARTOFBASICBLOCK(npc)) {
			end_basic_block();
			start_basic_block();
			stackno = STACKPOINTER(npc);
		}
	}
	assert(maxLocal + maxStack + maxTemp < MAXSLOT);

	/* Generate and link code */
	clen = CODEPC - codeblock;

	linkLabels();

	/* Note where it all is */
	meth->ncode = codeblock;
	meth->ncode_end = codeblock + clen;

	codeblock = CODEPC;

	/* Translate exception table and make it available */
	if (meth->exception_table != 0) {
		for (i = 0; i < meth->exception_table_len; i++) {
			e = &meth->exception_table[i];
			e->start_pc = codeInfo.perPC[e->start_pc].pc;
			e->end_pc = codeInfo.perPC[e->end_pc].pc;
			e->handler_pc = codeInfo.perPC[e->handler_pc].pc;
		}
	}

	/* Translate line numbers table */
	if (meth->lines != 0) {
		for (i = 0; i < meth->lines->length; i++) {
			meth->lines->entry[i].start_pc = codeInfo.perPC[meth->lines->entry[i].start_pc].pc;
		}
	}

MDBG(	printf("Translating %s.%s%s (%s) 0x%x\n", meth->class->name, meth->pair->s1, meth->pair->s2, isStatic ? "static" : "normal", meth->ncode);	)

	if (flag_jit) {
		tme = currentTime();
		jitStats.time += (tme - tms);
		jitStats.bytecodes += meth->codelen;
		printf("<JIT: %s.%s%s [time %dms (%dms)][codes %d (%d/s)] 0x%x>\n",
			CLASS_CNAME(meth->class),
			meth->name->data, meth->signature->data,
			(int)(tme - tms), jitStats.time,
			jitStats.bytecodes,
			(1000 * jitStats.bytecodes) / jitStats.time,
			meth->ncode);
		fflush(stdout);
	}

	_unlockMutex(&translatorMux);
}

/*
 * Start a new instruction.
 */
void
startInsn(uintp pc)
{
	/* Note PC value at start of this instruction */
	codeInfo.perPC[pc].pc = (uintp)CODEPC;
}

/*
 * Start a new basic block.
 */
void
startBlock(void)
{
	int i;

	/* Invalidate all slots - don't use clobberRegister which will
	 * flush them - we do not want to do that even if they are dirty.
	 * Also ignore any fixed slots.
	 */
	for (i = MAXSLOT - 1; i >= 0; i--) {
		if (cslotinfo[i].reg != 0 && cslotinfo[i].fixedreg == 0) {
			freeSlot(&cslotinfo[i]);
		}
	}
}

/*
 * End a basic block.
 */
void
endBlock(void)
{
	int i;

	/* Spill slots to memory which aren't fixed. */
	for (i = MAXSLOT - 1; i >= 0; i--) {
		if (cslotinfo[i].reg != 0 && cslotinfo[i].fixedreg == 0) {
			if (cslotinfo[i].reg->modified & rnowriteback) {
				cslotinfo[i].reg->modified &= ~rnowriteback;
			}
			else {
				spill(&cslotinfo[i]);
			}
		}
	}
}

/*
 * Fixup after a function call.
 */
void
fixupFunctionCall(void)
{
	int i;

	/* Invalidate all slots - don't use clobberRegister which will
	 * flush them - we do not want to do that even if they are dirty.
	 */
	for (i = MAXSLOT - 1; i >= 0; i--) {
		if (cslotinfo[i].reg != 0) {
			freeSlot(&cslotinfo[i]);
		}
	}
}

/*
 * Prepare to make a function call.
 */
void
prepareFunctionCall(void)
{
	int i;

	/* Spill all slots to memory (including fixed). */
	for (i = MAXSLOT - 1; i >= 0; i--) {
		if (cslotinfo[i].reg != 0) {
			if (cslotinfo[i].reg->modified & rnowriteback) {
				cslotinfo[i].reg->modified &= ~rnowriteback;
			}
			else {
				spill(&cslotinfo[i]);
			}
		}
	}
	if (activeInner.flow != 0) {
		activeInner.flow->cancel = true;
	}
}
